/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { isFlexTreeNode } from "../feature-libraries";
import { SchemaFactory } from "./schemaFactory";
/**
 * Extends SchemaFactory with utilities for recursive types.
 *
 * @remarks This is separated from SchemaFactory as these APIs are more experimental, and may be stabilized independently.
 *
 * @sealed @internal
 */
export class SchemaFactoryRecursive extends SchemaFactory {
    /**
     * For unknown reasons, recursive lists work better (compile in more cases)
     * if their constructor takes in an object with a member containing the iterable,
     * rather than taking the iterable as a parameter directly.
     *
     * This version of `list` leverages this fact, and has a constructor that requires its data be passed in like:
     * ```typescript
     * new MyRecursiveList({x: theData});
     * ```
     */
    arrayRecursive(name, allowedTypes) {
        class RecursiveArray extends this.namedArray(name, allowedTypes, true, false) {
            constructor(data) {
                if (isFlexTreeNode(data)) {
                    super(data);
                }
                else {
                    super(data.x);
                }
            }
        }
        return RecursiveArray;
    }
    /**
     * For unknown reasons, recursive maps work better (compile in more cases)
     * if their constructor does not take in the desired type.
     *
     * This version of `map` leverages this fact and takes in undefined instead.
     * Unfortunately this means all maps created this way must be created empty then filled later.
     * @privateRemarks
     * TODO:
     * Figure out a way to make recursive prefilled maps work.
     */
    mapRecursive(name, allowedTypes) {
        class MapSchema extends this.namedMap(name, allowedTypes, true, false) {
            constructor(data) {
                if (isFlexTreeNode(data)) {
                    super(data);
                }
                else {
                    super(new Map());
                }
            }
        }
        return MapSchema;
    }
}
//# sourceMappingURL=schemaFactoryRecursive.js.map